@@ -20,11 +20,14 @@
 */
 
 #include "config.h"
+#include <sys/types.h>
+#include <regex.h>
 #include <errno.h>
 #include <stdio.h>
 #include <string.h>
 #include <stdint.h>
 #include <stdlib.h>
+#include <stdarg.h>
 /* For error text */
 #include <libintl.h>
 #define _(String) gettext (String)
@@ -1039,3 +1042,737 @@ int ini_config_get_errors(struct ini_cfgobj *cfg_ctx,
     TRACE_FLOW_EXIT();
     return error;
 }
+
+int ini_rules_read_from_file(const char *filename,
+                             struct ini_cfgobj **_rules_obj)
+{
+    int ret;
+    struct ini_cfgfile *cfgfile = NULL;
+
+    if (_rules_obj == NULL) {
+        return EINVAL;
+    }
+
+    ret = ini_config_create(_rules_obj);
+    if (ret != EOK) {
+        return ret;
+    }
+
+    ret = ini_config_file_open(filename, 0, &cfgfile);
+    if (ret != EOK) {
+        goto done;
+    }
+
+    ret = ini_config_parse(cfgfile, 0, INI_MV1S_ALLOW, 0, *_rules_obj);
+    if (ret != EOK) {
+        goto done;
+    }
+
+done:
+    if (ret != EOK) {
+        ini_config_destroy(*_rules_obj);
+        *_rules_obj = NULL;
+    }
+
+    ini_config_file_destroy(cfgfile);
+    return ret;
+}
+
+/* This is used for testing only */
+static int ini_dummy_noerror(const char *rule_name,
+                             struct ini_cfgobj *rules_obj,
+                             struct ini_cfgobj *config_obj,
+                             struct ini_errobj *errobj,
+                             void **data)
+{
+    return 0;
+}
+
+/* This is used for testing only */
+static int ini_dummy_error(const char *rule_name,
+                           struct ini_cfgobj *rules_obj,
+                           struct ini_cfgobj *config_obj,
+                           struct ini_errobj *errobj,
+                           void **data)
+{
+    return ini_errobj_add_msg(errobj, "Error");
+}
+
+static int is_allowed_section(const char *tested_section,
+                              char **allowed_sections,
+                              size_t num_sec,
+                              regex_t *allowed_sections_re,
+                              size_t num_sec_re,
+                              int case_insensitive)
+{
+    int ret;
+    int i;
+
+    if (case_insensitive) {
+        for (i = 0; i < num_sec; i++) {
+            if (strcasecmp(tested_section, allowed_sections[i]) == 0) {
+                return 1;
+            }
+        }
+    } else { /* case sensitive */
+        for (i = 0; i < num_sec; i++) {
+            if (strcmp(tested_section, allowed_sections[i]) == 0) {
+                return 1;
+            }
+        }
+    }
+
+    for (i = 0; i < num_sec_re; i++) {
+        ret = regexec(&allowed_sections_re[i], tested_section, 0, NULL, 0);
+        if (ret == 0) {
+            return 1;
+        }
+    }
+
+    return 0;
+}
+
+static int ini_allowed_sections(const char *rule_name,
+                                struct ini_cfgobj *rules_obj,
+                                struct ini_cfgobj *config_obj,
+                                struct ini_errobj *errobj,
+                                void **data)
+{
+    struct value_obj *vo = NULL;
+    int ret;
+    char *regex_str = NULL;
+    char **allowed_sections = NULL;
+    char *insensitive_str;
+    char **cfg_sections = NULL;
+    int num_cfg_sections;
+    char **attributes = NULL;
+    int num_attributes;
+    size_t num_sec = 0;
+    size_t num_sec_re = 0;
+    regex_t *allowed_sections_re = NULL;
+    size_t buf_size;
+    int reg_err;
+    int is_allowed;
+    int case_insensitive = 0;
+    int regcomp_flags = REG_NOSUB;
+    int i;
+
+    /* Get number of 'section' and 'section_re' attributes
+     * in this rule */
+    attributes = ini_get_attribute_list(rules_obj,
+                                        rule_name,
+                                        &num_attributes,
+                                        NULL);
+    if (attributes == NULL) {
+        ret = ENOMEM;
+        goto done;
+    }
+
+    for (i = 0; i < num_attributes; i++) {
+        if (strcmp("section", attributes[i]) == 0) {
+            num_sec++;
+        }
+
+        if (strcmp("section_re", attributes[i]) == 0) {
+            num_sec_re++;
+        }
+    }
+
+    ini_free_attribute_list(attributes);
+
+    if (num_sec == 0 && num_sec_re == 0) {
+        /* This rule is empty. */
+        ret = ini_errobj_add_msg(errobj,
+                                 "No allowed sections specified. "
+                                 "Use 'section = default' to allow only "
+                                 "default section");
+        goto done;
+    }
+
+    ret = ini_get_config_valueobj(rule_name,
+                                  "case_insensitive",
+                                  rules_obj,
+                                  INI_GET_NEXT_VALUE,
+                                  &vo);
+    if (ret) {
+        goto done;
+    }
+
+    if (vo) {
+        insensitive_str = ini_get_string_config_value(vo, &ret);
+        if (ret) {
+            goto done;
+        }
+
+        if (strcasecmp(insensitive_str, "yes") == 0
+            || strcasecmp(insensitive_str, "true") == 0
+            || strcmp(insensitive_str, "1") == 0) {
+            case_insensitive = 1;
+            regcomp_flags |= REG_ICASE;
+        }
+
+        free(insensitive_str);
+    }
+
+    /* Create arrays for section_re regexes and section name
+     * strings. */
+    allowed_sections = calloc(num_sec + 1, sizeof(char *));
+    if (allowed_sections == NULL) {
+        ret = ENOMEM;
+        goto done;
+    }
+
+    allowed_sections_re = calloc(num_sec_re + 1, sizeof(regex_t));
+    if (allowed_sections_re == NULL) {
+        ret = ENOMEM;
+        goto done;
+    }
+
+    /* Get all allowed section names and store them to
+     * allowed_sections array */
+    for (i = 0; i < num_sec; i++) {
+        ret = ini_get_config_valueobj(rule_name,
+                                      "section",
+                                      rules_obj,
+                                      INI_GET_NEXT_VALUE,
+                                      &vo);
+        if (ret) {
+            goto done;
+        }
+
+        allowed_sections[i] = ini_get_string_config_value(vo, &ret);
+        if (ret) {
+            goto done;
+        }
+    }
+
+    /* Get all regular section_re regular expresions and
+     * store them to allowed_sections_re array */
+    for (i = 0; i < num_sec_re; i++) {
+        ret = ini_get_config_valueobj(rule_name,
+                                      "section_re",
+                                      rules_obj,
+                                      INI_GET_NEXT_VALUE,
+                                      &vo);
+        if (ret) {
+            goto done;
+        }
+
+        regex_str = ini_get_string_config_value(vo, &ret);
+        if (ret) {
+            goto done;
+        }
+
+        reg_err = regcomp(&allowed_sections_re[i], regex_str, regcomp_flags);
+        if (reg_err) {
+            char *err_str;
+
+            buf_size = regerror(reg_err, &allowed_sections_re[i], NULL, 0);
+            err_str = malloc(buf_size);
+            if (err_str == NULL) {
+                ret = ENOMEM;
+                goto done;
+            }
+
+            regerror(reg_err, &allowed_sections_re[i], err_str, buf_size);
+            ret = ini_errobj_add_msg(errobj,
+                                     "Validator failed to use regex [%s]:[%s]",
+                                     regex_str, err_str);
+            free(err_str);
+            ret = ret ? ret : EINVAL;
+            goto done;
+        }
+        free(regex_str);
+        regex_str = NULL;
+    }
+
+    /* Finally get list of all sections in configuration and
+     * check if they are matched by some string in allowed_sections
+     * or regex in allowed_sections_re */
+    cfg_sections = ini_get_section_list(config_obj, &num_cfg_sections, &ret);
+    if (ret != EOK) {
+        goto done;
+    }
+
+    for (i = 0; i < num_cfg_sections; i++) {
+        is_allowed = is_allowed_section(cfg_sections[i],
+                                        allowed_sections,
+                                        num_sec,
+                                        allowed_sections_re,
+                                        num_sec_re,
+                                        case_insensitive);
+        if (!is_allowed) {
+            ret = ini_errobj_add_msg(errobj,
+                                     "Section [%s] is not allowed. "
+                                     "Check for typos.",
+                                     cfg_sections[i]);
+            if (ret) {
+                goto done;
+            }
+        }
+    }
+
+    ret = EOK;
+done:
+    if (allowed_sections != NULL) {
+        for (i = 0; allowed_sections[i] != NULL; i++) {
+            free(allowed_sections[i]);
+        }
+        free(allowed_sections);
+    }
+    if (allowed_sections_re != NULL) {
+        for (i = 0; i < num_sec_re; i++) {
+            regfree(&allowed_sections_re[i]);
+        }
+        free(allowed_sections_re);
+    }
+    ini_free_section_list(cfg_sections);
+    free(regex_str);
+
+    return ret;
+}
+
+static int check_if_allowed(char *section, char *attr, char **allowed,
+                            int num_allowed, struct ini_errobj *errobj)
+{
+    int is_allowed = 0;
+    int ret;
+    int i;
+
+    for (i = 0; i < num_allowed; i++) {
+        if (strcmp(attr, allowed[i]) == 0) {
+            is_allowed = 1;
+            break;
+        }
+    }
+
+    if (!is_allowed) {
+        ret = ini_errobj_add_msg(errobj,
+                                 "Attribute '%s' is not allowed in "
+                                 "section '%s'. Check for typos.",
+                                 attr, section);
+        return ret;
+    }
+
+    return 0;
+}
+
+static int ini_allowed_options(const char *rule_name,
+                               struct ini_cfgobj *rules_obj,
+                               struct ini_cfgobj *config_obj,
+                               struct ini_errobj *errobj,
+                               void **data)
+{
+    struct value_obj *vo = NULL;
+    int ret;
+    char *section_regex;
+    int num_sections;
+    char **sections = NULL;
+    char **attributes = NULL;
+    int num_attributes;
+    int num_opts = 0;
+    int i;
+    int a;
+    regex_t preg;
+    size_t buf_size;
+    char *err_str = NULL;
+    int reg_err;
+    char **allowed = NULL;
+
+    /* Get section regex */
+    ret = ini_get_config_valueobj(rule_name,
+                                  "section_re",
+                                  rules_obj,
+                                  INI_GET_FIRST_VALUE,
+                                  &vo);
+    if (ret != 0) {
+        return ret;
+    }
+
+    if (vo == NULL) {
+        ret = ini_errobj_add_msg(errobj,
+                                 "Validator misses 'section_re' parameter");
+        if (ret) {
+            return ret;
+        }
+        return EINVAL;
+    }
+
+    section_regex = ini_get_string_config_value(vo, NULL);
+    if (section_regex == NULL || section_regex[0] == '\0') {
+        ret = ini_errobj_add_msg(errobj,
+                                 "Validator misses 'section_re' parameter");
+        if (ret) {
+            return ret;
+        }
+
+        free(section_regex);
+        return EINVAL;
+    }
+
+    /* compile the regular expression */
+    reg_err = regcomp(&preg, section_regex, REG_NOSUB);
+    if (reg_err) {
+        buf_size = regerror(reg_err, &preg, NULL, 0);
+        err_str = malloc(buf_size);
+        if (err_str == NULL) {
+            ret = ENOMEM;
+            goto done;
+        }
+
+        regerror(reg_err, &preg, err_str, buf_size);
+        ret = ini_errobj_add_msg(errobj,
+                                 "Cannot compile regular expression from "
+                                 "option 'section_re'. Error: '%s'", err_str);
+        ret = ret ? ret : EINVAL;
+        goto done;
+    }
+
+    /* Get all sections from config_obj */
+    sections = ini_get_section_list(config_obj, &num_sections, &ret);
+    if (ret != EOK) {
+        goto done;
+    }
+
+    /* Get number of 'option' attributes in this rule
+     * and create an array long enough to store them all */
+    attributes = ini_get_attribute_list(rules_obj,
+                                        rule_name,
+                                        &num_attributes,
+                                        NULL);
+    if (attributes == NULL) {
+        ret = ENOMEM;
+        goto done;
+    }
+
+    for (i = 0; i < num_attributes; i++) {
+        if (strcmp("option", attributes[i]) == 0) {
+            num_opts++;
+        }
+    }
+
+    ini_free_attribute_list(attributes);
+    attributes = NULL;
+
+    allowed = calloc(num_opts + 1, sizeof(char *));
+    if (allowed == NULL) {
+        ret = ENOMEM;
+        goto done;
+    }
+
+    for (i = 0; i < num_opts; i++) {
+        ret = ini_get_config_valueobj(rule_name,
+                                      "option",
+                                      rules_obj,
+                                      INI_GET_NEXT_VALUE,
+                                      &vo);
+        if (ret) {
+            goto done;
+        }
+
+        allowed[i] = ini_get_string_config_value(vo, &ret);
+        if (ret) {
+            goto done;
+        }
+    }
+
+    for (i = 0; i < num_sections; i++) {
+        if (regexec(&preg, sections[i], 0, NULL, 0) == 0) {
+            /* Regex matched section */
+            /* Get options from this section */
+            attributes = ini_get_attribute_list(config_obj,
+                                                sections[i],
+                                                &num_attributes,
+                                                NULL);
+            if (attributes == NULL) {
+                ret = ENOMEM;
+                goto done;
+            }
+
+            for (a = 0; a < num_attributes; a++) {
+                ret = check_if_allowed(sections[i], attributes[a], allowed,
+                                       num_opts, errobj);
+                if (ret != 0) {
+                    goto done;
+                }
+            }
+            ini_free_attribute_list(attributes);
+            attributes = NULL;
+        }
+    }
+
+    ret = 0;
+done:
+    if (allowed != NULL) {
+        for (i = 0; allowed[i] != NULL; i++) {
+            free(allowed[i]);
+        }
+        free(allowed);
+    }
+    ini_free_section_list(sections);
+    free(section_regex);
+    ini_free_attribute_list(attributes);
+    regfree(&preg);
+    free(err_str);
+    return ret;
+}
+
+static ini_validator_func *
+get_validator(char *validator_name,
+              struct ini_validator **validators)
+{
+    struct ini_validator *ext_validator;
+
+    /* First we check all internal validators */
+    if (strcmp(validator_name, "ini_dummy_noerror") == 0) {
+        return ini_dummy_noerror;
+    } else if (strcmp(validator_name, "ini_dummy_error") == 0) {
+        return ini_dummy_error;
+    } else if (strcmp(validator_name, "ini_allowed_options") == 0) {
+        return ini_allowed_options;
+    } else if (strcmp(validator_name, "ini_allowed_sections") == 0) {
+        return ini_allowed_sections;
+    }
+
+    if (validators == NULL) {
+        return NULL;
+    }
+
+    for (; *validators != NULL; ++validators) {
+        ext_validator = *validators;
+
+        /* Skip invalid external validator. Name is required */
+        if (ext_validator->name == NULL) {
+            continue;
+        }
+        if (strcmp(validator_name, ext_validator->name) == 0) {
+            return ext_validator->func;
+        }
+    }
+
+    return NULL;
+}
+
+int ini_rules_check(struct ini_cfgobj *rules_obj,
+                    struct ini_cfgobj *config_obj,
+                    struct ini_validator **extra_validators,
+                    struct ini_errobj *errobj)
+{
+    char **sections;
+    int ret;
+    int num_sections;
+    char *vname;
+    ini_validator_func *vfunc;
+    struct value_obj *vo = NULL;
+    struct ini_errobj *localerr = NULL;
+    int i;
+
+    /* Get all sections from the rules object */
+    sections = ini_get_section_list(rules_obj, &num_sections, &ret);
+    if (ret != EOK) {
+        return ret;
+    }
+
+    /* Now iterate through all the sections. If the section
+     * name begins with a prefix "rule/", then it is a rule
+     * name. */
+    for (i = 0; i < num_sections; i++) {
+        if (!strncmp(sections[i], "rule/", sizeof("rule/") - 1)) {
+            ret = ini_get_config_valueobj(sections[i],
+                                          "validator",
+                                          rules_obj,
+                                          INI_GET_FIRST_VALUE,
+                                          &vo);
+            if (ret != 0) {
+                /* Failed to get value object. This should not
+                 * happen. */
+                continue;
+            }
+
+            if (vo == NULL) {
+                ret = ini_errobj_add_msg(errobj,
+                                         "Rule '%s' has no validator.",
+                                         sections[i]);
+                if (ret != EOK) {
+                    return ret;
+                }
+                /* Skip problematic rule */
+                continue;
+            }
+
+            vname = ini_get_string_config_value(vo, NULL);
+            vfunc = get_validator(vname, extra_validators);
+            if (vfunc == NULL) {
+                ret = ini_errobj_add_msg(errobj,
+                                         "Rule '%s' uses unknown "
+                                         "validator '%s'.",
+                                         sections[i], vname);
+                if (ret != EOK) {
+                    goto done;
+                }
+                /* Skip problematic rule */
+                free(vname);
+                continue;
+            }
+            free(vname);
+
+            /* Do not pass global errobj to validators, they
+             * could corrupt it. Create local one for each
+             * validator. */
+            ret = ini_errobj_create(&localerr);
+            if (ret != EOK) {
+                goto done;
+            }
+
+            ret = vfunc(sections[i], rules_obj, config_obj, localerr, NULL);
+            if (ret != 0) {
+                /* Just report the error and continue normally,
+                 * maybe there are some errors in localerr */
+                ret = ini_errobj_add_msg(errobj,
+                                         "Rule '%s' returned error code '%d'",
+                                         sections[i], ret);
+                if (ret != EOK) {
+                    goto done;
+                }
+            }
+
+            /* Bad validator could destroy the localerr, check
+             * for NULL */
+            if (localerr == NULL) {
+                continue;
+            }
+
+            ini_errobj_reset(localerr);
+            while (!ini_errobj_no_more_msgs(localerr)) {
+                ret = ini_errobj_add_msg(errobj,
+                                         "[%s]: %s",
+                                         sections[i],
+                                         ini_errobj_get_msg(localerr));
+                if (ret != EOK) {
+                    goto done;
+                }
+                ini_errobj_next(localerr);
+            }
+
+            ini_errobj_destroy(&localerr);
+        }
+    }
+
+    ret = EOK;
+done:
+    ini_free_section_list(sections);
+    ini_errobj_destroy(&localerr);
+    return ret;
+}
+
+/* This is just convenience function, so that
+ * we manipulate with ini_rules_* functions. */
+void ini_rules_destroy(struct ini_cfgobj *rules)
+{
+    ini_config_destroy(rules);
+}
+
+int ini_errobj_create(struct ini_errobj **_errobj)
+{
+    struct ini_errobj *new_errobj = NULL;
+
+    if (_errobj == NULL) {
+        return EINVAL;
+    }
+
+    new_errobj = calloc(1, sizeof(struct ini_errobj));
+    if (new_errobj == NULL) {
+        return ENOMEM;
+    }
+
+    *_errobj = new_errobj;
+    return EOK;
+}
+
+void ini_errobj_destroy(struct ini_errobj **errobj)
+{
+    struct ini_errmsg *to_remove;
+
+    if (errobj == NULL || *errobj == NULL) {
+        return;
+    }
+
+    while ((*errobj)->first_msg) {
+        to_remove = (*errobj)->first_msg;
+        (*errobj)->first_msg = (*errobj)->first_msg->next;
+        free(to_remove->str);
+        free(to_remove);
+    }
+
+    free(*errobj);
+    *errobj = NULL;
+}
+
+int ini_errobj_add_msg(struct ini_errobj *errobj, const char *format, ...)
+{
+    int ret;
+    va_list args;
+    struct ini_errmsg *new;
+
+    new = calloc(1, sizeof(struct ini_errmsg));
+    if (new == NULL) {
+        return ENOMEM;
+    }
+
+    va_start(args, format);
+    ret = vasprintf(&new->str, format, args);
+    va_end(args);
+    if (ret == -1) {
+        free(new);
+        return ENOMEM;
+    }
+
+    if (errobj->count == 0) {
+        /* First addition to the list, all pointers are NULL */
+        errobj->first_msg = new;
+        errobj->last_msg = new;
+        errobj->cur_msg = new;
+        errobj->count++;
+    } else {
+        errobj->last_msg->next = new;
+        errobj->last_msg = errobj->last_msg->next;
+        errobj->count++;
+    }
+
+    return EOK;
+}
+
+void ini_errobj_reset(struct ini_errobj *errobj)
+{
+    errobj->cur_msg = errobj->first_msg;
+}
+
+const char *ini_errobj_get_msg(struct ini_errobj *errobj)
+{
+    if (errobj->cur_msg != NULL) {
+        return errobj->cur_msg->str;
+    }
+
+    /* Should this be allowed? */
+    return NULL;
+}
+
+void ini_errobj_next(struct ini_errobj *errobj)
+{
+    if (errobj->cur_msg != NULL) {
+        errobj->cur_msg = errobj->cur_msg->next;
+    }
+
+    /* If we can not move next, just return */
+    return;
+}
+
+int ini_errobj_no_more_msgs(struct ini_errobj *errobj)
+{
+    return errobj->cur_msg == NULL;
+}
+
+size_t ini_errobj_count(struct ini_errobj *errobj)
+{
+    return errobj->count;
+}
