@@ -32,6 +32,14 @@
 #include "simplebuffer.h"
 #include "ini_valueobj.h"
 
+#ifndef DING_ATTR_FORMAT
+#  if ((__GNUC__ > 3) || ((__GNUC__ == 3) && (__GNUC_MINOR__ >= 3)))
+#    define DING_ATTR_FORMAT(fmt, args) __attribute__((__format__(__printf__, fmt, args)))
+#  else
+#    define DING_ATTR_FORMAT(fmt, args)
+#  endif
+#endif
+
 /** @mainpage The INI configuration interface
  *
  * The goal of the this interface is to allow applications
@@ -363,6 +371,8 @@ enum ERR_PARSE {
 #define INI_PARSE_NOTAB            0x0004
 /** @brief Do not allow C-style comments */
 #define INI_PARSE_NO_C_COMMENTS    0x0008
+/** @brief Skip lines that are not KVPs */
+#define INI_PARSE_IGNORE_NON_KVP    0x0010
 
 /**
  * @}
@@ -1005,7 +1015,8 @@ int ini_config_merge(struct ini_cfgobj *first,
  *
  * Function merges the main configuration file
  * with the configuration file snippets
- * read from a specified directory.
+ * read from a specified directory
+ * ordered according to the locale setting.
  *
  * @param[in]  base_cfg         A configuration object
  *                              that will be augmented.
@@ -2045,6 +2056,185 @@ void ini_free_long_config_array(long *array);
  */
 void ini_free_double_config_array(double *array);
 
+/** @brief Structure that holds error messages
+ *  generated by validators.
+ */
+struct ini_errobj;
+
+/**
+ * @brief Create structure to hold error messages.
+ *
+ * This function initiates structure that can be used to
+ * hold error messages from generators. To add messages to
+ * the structure use \ref ini_errobj_add_msg.
+ *
+ * @param[out] errobj         container for errors.
+ *
+ * @return Zero on success, nonzero value in case of error.
+ */
+int ini_errobj_create(struct ini_errobj **_errobj);
+
+/**
+ * @brief Free structure that holds error messages.
+ *
+ * This function is used to free structure
+ * previously created by \ref ini_errobj_create.
+ *
+ * @param[in] errobj         container for errors.
+ */
+void ini_errobj_destroy(struct ini_errobj **errobj);
+
+/**
+ * @brief Add new printf formated message to errobj.
+ *
+ * This function initiates structure that can be used to
+ * hold error messages from generators. To add messages to
+ * the structure use \ref ini_errobj_add_msg.
+ *
+ * @param[in] errobj         container for errors previously
+ *                           created by \ref ini_errobj_create.
+ * @param[in] format         printf format string
+ *
+ * @return Zero on success, nonzero value in case of error.
+ */
+int ini_errobj_add_msg(struct ini_errobj *errobj,
+                       const char *format, ...)
+                       DING_ATTR_FORMAT(2, 3);
+
+/**
+ * @brief Reset iterator in errobj.
+ *
+ * After calling this function, the iterator in errobj
+ * will point to the first error message. Use this if
+ * you need to accesss the list multiple times in a loop.
+ *
+ * @param[in] errobj         container for errors previously
+ *                           created by \ref ini_errobj_create.
+ */
+void ini_errobj_reset(struct ini_errobj *errobj);
+
+/**
+ * @brief Get pointer to current message in errobj.
+ *
+ * This function returns pointer to current message
+ * pointed by the internal iterator. The returned string can
+ * not be changed and will point to valid data only
+ * until \ref ini_errobj_destroy is called.
+ *
+ * @param[in] errobj         container for errors previously
+ *                           created by \ref ini_errobj_create.
+ * @return String inside the errobj structure. String
+ * is valid until errobj is destroyed.
+ */
+const char *ini_errobj_get_msg(struct ini_errobj *errobj);
+
+/**
+ * @brief Move to the next message in errobj.
+ *
+ * This function moves the internal iterator of errobj
+ * to the next message in list.
+ *
+ * @param[in] errobj         container for errors previously
+ *                           created by \ref ini_errobj_create.
+ */
+void ini_errobj_next(struct ini_errobj *errobj);
+
+/**
+ * @brief Check if errobj has more messages.
+ *
+ * This function returns true if errobj's internal iterator
+ * reached end of list and no longer points to a message
+ *
+ * @param[in] errobj         container for errors previously
+ *                           created by \ref ini_errobj_create.
+ * @return True if internal iterator reached end of list.
+ */
+int ini_errobj_no_more_msgs(struct ini_errobj *errobj);
+
+/**
+ * @brief Return number of messages in errobj
+ *
+ * This function returns number of messages inside errobj
+ *
+ * @param[in] errobj         container for errors previously
+ *                           created by \ref ini_errobj_create.
+ * @return Number of messages stored in errobj.
+ */
+size_t ini_errobj_count(struct ini_errobj *errobj);
+
+typedef int (ini_validator_func)(const char *rule_name,
+                                 struct ini_cfgobj *rules_obj,
+                                 struct ini_cfgobj *config_obj,
+                                 struct ini_errobj *errobj,
+                                 void **data);
+
+typedef int (ini_schema_validator_func)(const char *rule_name,
+                                        struct ini_cfgobj *rules_obj,
+                                        struct ini_errobj *errobj,
+                                        void **data);
+
+/** @brief Structure used to define application specific
+ * (external to libini) validator
+ */
+struct ini_validator {
+    const char *name;
+    ini_validator_func *func;
+    /* currently unused, for future expansion */
+    ini_schema_validator_func *schema_validator;
+};
+
+/**
+ * @brief Read rules from INI file
+ *
+ * This function is used to read rules from INI file
+ * and store them in config object. This special
+ * config object is passed to \ref ini_rules_check
+ * together with config object representing the
+ * configuration that will be checked.
+ *
+ * @param[in] filename       Name of file with rules
+ * @param[out] _rules_obj    config object representing the rules
+ * @return Zero on success. Non zero value on error.
+ */
+int ini_rules_read_from_file(const char *filename,
+                             struct ini_cfgobj **_rules_obj);
+
+/**
+ * @brief Check configuration file using rules
+ *
+ * This function is used to check if configuration
+ * file applies to rules previously loaded by
+ * \ref ini_rules_read_from_file. Any errors
+ * detected in the configuration are stored in the
+ * errobj structure. Error code returned by this
+ * function indicates some internal error with
+ * validators or memory allocation error (not
+ * rule violation).
+ *
+ * @param[in] rules_obj            config object representing the rules
+ * @param[in] config_obj           config object representing the
+ *                                 configuration
+ * @param[in] extra_validators     NULL terminated array of external
+ *                                 validators. Can be NULL if no external
+ *                                 validators are used.
+ *
+ * @param[in] errobj               errobj to store generated errors
+ *                                 from validators.
+ *
+ * @return Zero on success. Non zero value on error.
+ */
+int ini_rules_check(struct ini_cfgobj *rules_obj,
+                    struct ini_cfgobj *config_obj,
+                    struct ini_validator **extra_validators,
+                    struct ini_errobj *errobj);
+
+/**
+ * @brief Free the rules
+ *
+ * This function is just wrapper around ini_config_destroy
+ */
+void ini_rules_destroy(struct ini_cfgobj *ini_config);
+
 /**
  * @}
  */
